;Simple clock with ATmega13 and LED matrix using Charlieplexing Plus,
;mains frequency timebase (50 Hz)
;ATtiny13 @ 9.6 MHz (int. osc), Fuses: low = 0xE2, high = 0xFA [RESET disabled]
;   start-up clock frequency: 1.2 MHz (9.6 MHz / 8 - CLKDIV fuse enabled)
;   operation clock frequency: 4.8 MHz (9.6 MHz / 2 - set by clock prescaler)
;   can't use 4.8 MHz fuse setting as automatic calibration is only for 9.6 MHz
;
;PORTS
;  PB0 = LED matrix A - 18R series resistor
;  PB1 = LED matrix B - 18R series resistor
;  PB2 = LED matrix C
;  PB3 = LED matrix D
;  PB4 = LED matrix E
;  PB5 = 50 Hz input - RESET pin disabled!
;
;LED MATRIX
;  Digits: 1 2 : 3 4
;
;  Two LEDs in series from pin to power rail:
;     Pin | Rail | Digit/Segment
;    -----+------+---------------
;      A  | VCC  | 1a
;      A  | GND  | 1d
;      B  | VCC  | 2a
;      B  | GND  | 2d
;      C  | VCC  | colon
;      C  | GND  | push-button with 1k series resistor
;      D  | VCC  | 3a
;      D  | GND  | 3d
;      E  | VCC  | 4a
;      E  | GND  | 4d
;
;  Two LEDs in parallel between two pins:
;       | ,--------- Anode ---------,
;       |  A  |  B  |  C  |  D  |  E
;    ---+-----+-----+-----+-----+-----
;     A | --  | 1b  | 2b  | 3b  | 4b
;     B | 1f  | --  | 1c  | 2c  | 2g
;     C | 2f  | 1e  | --  | 3c  | 4c
;     D | 3f  | 2e  | 3e  | --  | 4g
;     E | 4f  | 1g  | 4e  | 3g  | --
;
;VERSIONS
;  2017-10-22  Arne Rossius
;   * first version
;  2017-11-02  Arne Rossius
;   * improved time correction for LEDs in series
;  2018-12-04  Arne Rossius
;   * switched from 4.8 MHz to 9.6 MHz clock for better calibration
;   * decreased zero crossing hold-off to 80% of nominal

;===============================================================================

;clock settings
.equ FLASHING_COLON = 1 ;colon flashing frequency (0=static, 1=1 Hz, 2=0.5 Hz)
.equ ZERO_BLANKING = 1 ;blank 1st digit zero (0 = no, 1 = yes)
.equ MAINS_60HZ = 0 ;use 60 Hz mains frequency (0 = 50 Hz, 1 = 60 Hz)

;button press durations (unit: mux = 208.333 Hz => 4.8 ms)
.equ	BUTTON_SHORT_TIME = 3 ;~ 14 ms
.equ	BUTTON_LONG_TIME = 104 ;~500 ms

;display flashing speed (unit: mux = 208.333 Hz => 4.8 ms)
.equ	FLASH_SPEED = 208 ;~1 Hz

;display characters (LSB/bit 0 = segment a ... bit 6 = segment g)
.equ DIGIT_0 = 0x3F
.equ DIGIT_1 = 0x06
.equ DIGIT_2 = 0x5B
.equ DIGIT_3 = 0x4F
.equ DIGIT_4 = 0x66
.equ DIGIT_5 = 0x6D
.equ DIGIT_6 = 0x7D
.equ DIGIT_7 = 0x07
.equ DIGIT_8 = 0x7F
.equ DIGIT_9 = 0x6F
.equ DIGIT_COLON = 0x80 ;only on digit 2
.equ DIGIT_BLANK = 0x00

;mains zero crossing detection hold-off (unit: 4.8 MHz / 256 => 53.333 us)
.if (MAINS_60HZ)
  .equ	AC_TIMEOUT_VALUE = 250 ;13+1/3 ms (period: 16+2/3 ms)
.else
  .equ	AC_TIMEOUT_VALUE = 300 ;16 ms (period: 20 ms)
.endif
	
;===============================================================================

.include "../include/tn13def.inc"

.undef XL
.undef XH
.undef YL
.undef YH

.def	zero = R0

.def	ac_timeout_L = R12
.def	ac_timeout_H = R13
.def	minute = R14
.def	hour = R15

.def	temp = R16
.def	temp2 = R17
.def	subsec = R18
.def	second = R19
.def	digit1 = R20
.def	digit2 = R21
.def	digit3 = R22
.def	digit4 = R23
.def	mux = R24
.def	button_cnt = R25
.def	flash = R26
.def	set_digit = R27

.def	flags = R29
	.equ	fButtonShort = 0
	.equ	fButtonLong = 1
	.equ	fTimeSet = 2
	.equ	fAcState = 3

;R31:R30: flash pointer for multiplexing interrupt

;===============================================================================

.cseg
.org 0x000
	rjmp	reset

.org OVF0addr
	;timer 0 overflow interrupt: multiplex display
	;  mux =  0~19: segments between pins: 4*256 = 1024 cycles each
	;  mux = 20/21: segments to VCC/GND: 5*256 = 1280 cycles each
	;total: 23040 cycles => 208.333 Hz multiplexing @ 4.8 MHz
	push	temp
	push	temp2
	in	temp, SREG
	push	temp
	
	inc	mux
	cpi	mux, 22
	brlo	PC+2
	ldi	mux, 0
	cpi	mux, 20
	breq	mux_pos
	cpi	mux, 21
	breq	mux_neg
	clr	ZH
	mov	ZL, mux
	lsl	ZL
	lsl	ZL
	subi	ZL, -(mux_table)
	clr	temp
	ijmp
mux_output:
	and	temp2, temp
	out	DDRB, zero
	out	PORTB, temp2
	out	DDRB, temp
	ldi	temp, -4
	
mux_return:
	;return
	out	TCNT0, temp
	neg	temp
	sub	ac_timeout_L, temp
	sbc	ac_timeout_H, zero
	brcc	mux_ac_end
	clr	ac_timeout_L
	clr	ac_timeout_H
mux_ac_end:
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti
	
mux_pos:
	;A+, B+, D+, E+ - all digits segment d
	clr	temp
	bst	digit1, 3
	bld	temp, 0
	bst	digit2, 3
	bld	temp, 1
	bst	digit3, 3
	bld	temp, 3
	bst	digit4, 3
	bld	temp, 4
	out	DDRB, zero
	out	PORTB, temp
	out	DDRB, temp
	;flashing digit counter
	inc	flash
	cpi	flash, FLASH_SPEED
	brlo	flash_end
	clr	flash
flash_end:
	ldi	temp, -5
	rjmp	mux_return
	
mux_neg:
	;poll push-button
	ldi	temp, 0x04
	out	DDRB, zero
	out	PORTB, temp ;enable pull-up for push-button
	ldi	temp, 10
	dec	temp
	brne	PC-1
	sbic	PINB, 2
	rjmp	mux_button_released
	sbrs	button_cnt, 7
	inc	button_cnt
	cpi	button_cnt, BUTTON_LONG_TIME
	brne	mux_button_end
	ori	flags, 1<<fButtonLong
	rjmp	mux_button_end
mux_button_released:
	cpi	button_cnt, BUTTON_SHORT_TIME
	brlo	mux_button_short_end
	cpi	button_cnt, BUTTON_LONG_TIME
	brsh	mux_button_short_end
	ori	flags, 1<<fButtonShort
mux_button_short_end:
	clr	button_cnt
mux_button_end:
	;A-, B-, C-, D-, E- - all digits segment a & colon
	clr	temp
	bst	digit1, 0
	bld	temp, 0
	bst	digit2, 0
	bld	temp, 1
	bst	digit2, 7 ;colon
	bld	temp, 2
	bst	digit3, 0
	bld	temp, 3
	bst	digit4, 0
	bld	temp, 4
	out	DDRB, zero
	out	PORTB, zero
	out	DDRB, temp
	ldi	temp, -5
	rjmp	mux_return

mux_table:
	;0: AB - digit 1 segment f
	ldi	temp2, 0x01
	sbrc	digit1, 5
	ldi	temp, 0x03
	rjmp	mux_output
	;1: AC - digit 2 segment f
	ldi	temp2, 0x01
	sbrc	digit2, 5
	ldi	temp, 0x05
	rjmp	mux_output
	;2: AD - digit 3 segment f
	ldi	temp2, 0x01
	sbrc	digit3, 5
	ldi	temp, 0x09
	rjmp	mux_output
	;3: AE - digit 4 segment f
	ldi	temp2, 0x01
	sbrc	digit4, 5
	ldi	temp, 0x11
	rjmp	mux_output
	;4: BA - digit 1 segment b
	ldi	temp2, 0x02
	sbrc	digit1, 1
	ldi	temp, 0x03
	rjmp	mux_output
	;5: BC - digit 1 segment e
	ldi	temp2, 0x02
	sbrc	digit1, 4
	ldi	temp, 0x06
	rjmp	mux_output
	;6: BD - digit 2 segment e
	ldi	temp2, 0x02
	sbrc	digit2, 4
	ldi	temp, 0x0A
	rjmp	mux_output
	;7: BE - digit 1 segment g
	ldi	temp2, 0x02
	sbrc	digit1, 6
	ldi	temp, 0x12
	rjmp	mux_output
	;8: CA - digit 2 segment b
	ldi	temp2, 0x04
	sbrc	digit2, 1
	ldi	temp, 0x05
	rjmp	mux_output
	;9: CB - digit 1 segment c
	ldi	temp2, 0x04
	sbrc	digit1, 2
	ldi	temp, 0x06
	rjmp	mux_output
	;10: CD - digit 3 segment e
	ldi	temp2, 0x04
	sbrc	digit3, 4
	ldi	temp, 0x0C
	rjmp	mux_output
	;11: CE - digit 4 segment e
	ldi	temp2, 0x04
	sbrc	digit4, 4
	ldi	temp, 0x14
	rjmp	mux_output
	;12: DA - digit 3 segment b
	ldi	temp2, 0x08
	sbrc	digit3, 1
	ldi	temp, 0x09
	rjmp	mux_output
	;13: DB - digit 2 segment c
	ldi	temp2, 0x08
	sbrc	digit2, 2
	ldi	temp, 0x0A
	rjmp	mux_output
	;14: DC - digit 3 segment c
	ldi	temp2, 0x08
	sbrc	digit3, 2
	ldi	temp, 0x0C
	rjmp	mux_output
	;15: DE - digit 3 segment g
	ldi	temp2, 0x08
	sbrc	digit3, 6
	ldi	temp, 0x18
	rjmp	mux_output
	;16: EA - digit 4 segment b
	ldi	temp2, 0x10
	sbrc	digit4, 1
	ldi	temp, 0x11
	rjmp	mux_output
	;17: EB - digit 2 segment g
	ldi	temp2, 0x10
	sbrc	digit2, 6
	ldi	temp, 0x12
	rjmp	mux_output
	;18: EC - digit 4 segment c
	ldi	temp2, 0x10
	sbrc	digit4, 2
	ldi	temp, 0x14
	rjmp	mux_output
	;19: ED - digit 4 segment g
	ldi	temp2, 0x10
	sbrc	digit4, 6
	ldi	temp, 0x18
	rjmp	mux_output
mux_table_end:
.if (mux_table_end >= 0x100)
  .error "Multiplex table must end before word address 0x100"
.endif
	
;===============================================================================

reset:
	;ports
	ldi	temp, 0x00
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTB, temp
	ldi	temp, 0x1B ;disable digital input buffers on LED matrix pins
	out	DIDR0, temp
	
	;disable analog comparator (reduce current consumption)
	ldi	temp, 1<<ACD
	out	ACSR, temp
	
	;stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;set clock prescaler for 4.8 MHz clock
	ldi	temp, 1<<CLKPCE
	out	CLKPR, temp
	ldi	temp, 0x01 ;Clk/2  =>  9.6 MHz / 2 = 4.8 MHz
	out	CLKPR, temp
	
	;init timer 0 for multiplexing
	ldi	temp, 0x04 ;Clk/256
	out	TCCR0B, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK0, temp
	ldi	temp, -1
	out	TCNT0, temp
	
	;registers
	clr	zero
	ldi	mux, -1
	clr	button_cnt
	clr	flags
	clr	subsec
	clr	second
	clr	minute
	clr	hour
	clr	digit1
	clr	digit2
	clr	digit3
	clr	digit4
	clr	flash
	clr	set_digit
	clr	ac_timeout_L
	clr	ac_timeout_H
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	;check AC input
	cp	ac_timeout_L, zero
	cpc	ac_timeout_H, zero
	brne	ac_end
	sbrc	flags, fAcState
	rjmp	ac_falling
	;waiting for rising edge
	sbis	PINB, 5
	rjmp	ac_end
	ori	flags, 1<<fAcState
	ldi	temp, LOW(AC_TIMEOUT_VALUE)
	ldi	temp2, HIGH(AC_TIMEOUT_VALUE)
	movw	ac_timeout_H:ac_timeout_L, temp2:temp
	;increment time
	inc	subsec
.if (MAINS_60HZ)
	cpi	subsec, 60
.else
	cpi	subsec, 50
.endif
	brlo	ac_end
	clr	subsec
	inc	second ;second
	cpi	second, 60
	brlo	ac_end
	clr	second
	mov	temp, minute ;minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x60
	brlo	ac_end
	clr	minute
	mov	temp, hour ;hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	ac_end
	clr	hour
	rjmp	ac_end
ac_falling:
	;waiting for falling edge
	sbic	PINB, 5
	rjmp	ac_end
	andi	flags, ~(1<<fAcState)
ac_end:

	;check buttons
	sbrc	flags, fButtonShort
	rcall	button_short
	sbrc	flags, fButtonLong
	rcall	button_long
	
	;prepare digit data for hours
	mov	temp, hour
	rcall	convert
.if (ZERO_BLANKING)
	cpi	temp, DIGIT_0
	brne	PC+2
	ldi	temp, DIGIT_BLANK
.endif
	;flashing digits when setting time
	cpi	set_digit, 1
	sbrc	flags, fTimeSet
	brne	set_h_end
	cpi	flash, FLASH_SPEED/2
	brlo	set_h_end
	clr	temp
	clr	temp2
set_h_end:
.if (FLASHING_COLON == 1) ;1 Hz
  .if (MAINS_60HZ)
	cpi	subsec, 60/2
	brsh	PC+2
  .else
	cpi	subsec, 50/2
	brsh	PC+2
  .endif
.elif (FLASHING_COLON == 2) ;0.5 Hz
	sbrs	second, 0
.endif ;else: static (always on)
	ori	temp2, DIGIT_COLON
	mov	digit1, temp
	mov	digit2, temp2
	
	;prepare digit data for minutes
	mov	temp, minute
	rcall	convert
	;flashing digits when setting time
	cpi	set_digit, 2
	sbrc	flags, fTimeSet
	brne	set_m10_end
	cpi	flash, FLASH_SPEED/2
	brlo	set_m10_end
	clr	temp
set_m10_end:
	;flashing digits when setting time
	cpi	set_digit, 3
	sbrc	flags, fTimeSet
	brne	set_m1_end
	cpi	flash, FLASH_SPEED/2
	brlo	set_m1_end
	clr	temp2
set_m1_end:
	mov	digit3, temp
	mov	digit4, temp2
	
	rjmp	loop
	
	
;===============================================================================

button_short:
	;short button press
	andi	flags, ~(1<<fButtonShort)
	ldi	flash, 0
	cpi	set_digit, 1
	breq	adj_h1
	cpi	set_digit, 2
	breq	adj_m10
	cpi	set_digit, 3
	breq	adj_m1
	ret
	
adj_h1:
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x24
	brlo	PC+2
	clr	hour
	ret

adj_m10:
	clr	subsec
	clr	second
	mov	temp, minute
	andi	temp, 0xF0
	subi	temp, -0x10
	cpi	temp, 0x60
	brlo	PC+2
	subi	temp, 0x60
	mov	minute, temp
	ret
	
adj_m1:
	clr	subsec
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x60
	brlo	PC+2
	clr	minute
	ret
	
;--------------------
	
button_long:
	;long button press
	andi	flags, ~(1<<fButtonLong)
	ori	flags, 1<<fTimeSet
	inc	set_digit
	andi	set_digit, 0x03
	ldi	flash, FLASH_SPEED/2 ;digit off immediately
	ret

;===============================================================================

convert:
	;convert 'temp' to two 7-segment digits 'temp' and 'temp2'
	mov	temp2, temp
	andi	temp2, 0x0F
	swap	temp
	andi	temp, 0x0F
	ldi	ZL, LOW(convert_table*2)
	ldi	ZH, HIGH(convert_table*2)
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z
	ldi	ZL, LOW(convert_table*2)
	ldi	ZH, HIGH(convert_table*2)
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	lpm	temp2, Z
	ret

convert_table:
	;7-segment representations of 0x0 to 0xF (0xA to 0xF = blank)
	.db DIGIT_0, DIGIT_1, DIGIT_2, DIGIT_3
	.db DIGIT_4, DIGIT_5, DIGIT_6, DIGIT_7
	.db DIGIT_8, DIGIT_9, DIGIT_BLANK, DIGIT_BLANK
	.db DIGIT_BLANK, DIGIT_BLANK, DIGIT_BLANK, DIGIT_BLANK

;--------------------

bcdinc:
	;increment packed BCD number
	subi	temp, -7
	brhc	PC+2
	subi	temp, 6
	ret
